@@ -7,9 +7,10 @@ EOF
 
 _tmpdir="${BUILDDIR}/tmp"
 
-_packages="$(config_value regress-packages 2>/dev/null || :)"
-for _p in $_packages; do
-	if ! PKG_PATH='' pkg_info "$_p" >/dev/null; then
+{ config_value regress-packages 2>/dev/null || :; } |
+xargs printf '%s\n' | sort | uniq |
+while read -r _p; do
+	if ! PKG_PATH='' pkg_info "$_p" >/dev/null 2>&1; then
 		echo "$_p" >>"${_tmpdir}/packages"
 	fi
 done
