@@ -1,9 +1,9 @@
 . "${EXECDIR}/util.sh"
 
 config_load <<'EOF'
-BUILDDIR="${builddir}"
 DESTDIR="${destdir}"
 BSDSRCDIR="${bsd-srcdir}"
+RELEASEDIR="${bsd-reldir}"; export RELEASEDIR
 EOF
 
 chown build "$DESTDIR"
@@ -13,14 +13,14 @@ DESTDIR="${DESTDIR}/src"; export DESTDIR
 mkdir -p "$DESTDIR"
 cleandir "$DESTDIR"
 
-RELEASEDIR="$(release_dir "$BUILDDIR")"; export RELEASEDIR
 mkdir -p "$RELEASEDIR"
 cleandir "$RELEASEDIR"
 chown build "$RELEASEDIR"
 chmod 755 "$RELEASEDIR"
 
-# Remove vnd devices in case of resuming.
-make -C "${BSDSRCDIR}/distrib/$(machine)" unconfig
+# Remove vnd devices in case of resuming. As some architectures does not support
+# this make target, ignore errors.
+make -C "${BSDSRCDIR}/distrib/$(machine)" unconfig || :
 
 cd "${BSDSRCDIR}/etc"
 make release
