@@ -6,6 +6,34 @@ regress_config_load() {
 	unset MAKEFLAGS
 }
 
+# regress_duration_total -s steps
+#
+# Calculate the total duration. Since robsd-regress runs steps in parallel, the
+# accumulated step duration cannot be used. Instead, favor the wall clock delta
+# between the last and first step.
+regress_duration_total() {
+	local _t0=0
+	local _t1=0
+	local _steps
+
+	while [ $# -gt 0 ]; do
+		case "$1" in
+		-s)	shift; _steps="$1";;
+		*)	break;;
+		esac
+		shift
+	done
+	: "${_steps:?}"
+
+	if step_eval 1 "$_steps" 2>/dev/null; then
+		_t0="$(step_value time)"
+	fi
+	if step_eval -1 "$_steps" 2>/dev/null; then
+		_t1="$(step_value time)"
+	fi
+	echo "$((_t1 - _t0))"
+}
+
 # regress_failed step-log
 #
 # Exits zero if the given regress step log indicates failure.
@@ -36,83 +64,6 @@ regress_makefile() {
 	fi
 }
 
-# regress_report_log -e step-exit -n step-name -l step-log -t tmp-dir
-#
-# Get an excerpt of the given step log.
-regress_report_log() {
-	local _name
-	local _log
-	local _quiet=""
-
-	while [ $# -gt 0 ]; do
-		case "$1" in
-		-e)	shift;;
-		-n)	shift; _name="$1";;
-		-l)	shift; _log="$1";;
-		-t)	shift;;
-		*)	break;;
-		esac
-		shift
-	done
-	: "${_log:?}"
-	: "${_name:?}"
-
-	regress_quiet "$_name" || _quiet="yes"
-	regress_log -EFP ${_quiet:+-SX} "$_log" || tail "$_log"
-}
-
-# regress_report_skip -b build-dir -n step-name -l step-log -t tmp-dir
-#
-# Exits zero if the given step should not be included in the report.
-regress_report_skip() {
-	local _log
-	local _name
-	local _tmpdir
-
-	while [ $# -gt 0 ]; do
-		case "$1" in
-		-b)	shift;;
-		-l)	shift; _log="$1";;
-		-n)	shift; _name="$1";;
-		-t)	shift; _tmpdir="$1";;
-		*)	break;;
-		esac
-		shift
-	done
-	: "${_log:?}"
-	: "${_name:?}"
-
-	# Do not skip if one or many tests where skipped.
-	if ! regress_quiet "$_name" && regress_log -SXn "$_log"; then
-		return 1
-	fi
-	return 0
-}
-
-# regress_report_status -s steps
-#
-# Get the report status subject.
-regress_report_status() {
-	local _n
-	local _steps
-
-	while [ $# -gt 0 ]; do
-		case "$1" in
-		-s)	shift; _steps="$1";;
-		*)	break;;
-		esac
-		shift
-	done
-	: "${_steps:?}"
-
-	_n="$(step_failures "$_steps")"
-	if [ "$_n" -gt 1 ]; then
-		echo "${_n} failures"
-	elif [ "$_n" -gt 0 ]; then
-		echo "${_n} failure"
-	fi
-}
-
 # regress_root test
 #
 # Exits zero if the given regress test must be executed as root.
@@ -123,17 +74,6 @@ regress_root() {
 	config_value "regress-${_test}-root" >/dev/null 2>&1
 }
 
-# regress_quiet test
-#
-# Exits zero if the given regress test should be omitted from the report even if
-# some tests where skipped.
-regress_quiet() {
-	local _test
-
-	_test="$1"; : "${_test:?}"
-	config_value "regress-${_test}-quiet" >/dev/null 2>&1
-}
-
 # regress_step_after -b build-dir -e step-exit -n step-name
 #
 # After step hook, exits 0 if we can continue.
@@ -161,22 +101,14 @@ regress_step_after() {
 	return "$_exit"
 }
 
-# regress_steps
+# regress_step_parallel step-name
 #
-# Get the step names in order of execution.
-regress_steps() {
-	xargs printf '%s\n' <<-EOF
-	env
-	pkg-add
-	cvs
-	patch
-	obj
-	mount
-	$(config_value regress)
-	umount
-	revert
-	pkg-del
-	dmesg
-	end
-	EOF
+# Exits zero if the step can be executed in parallel.
+regress_step_parallel() {
+	local _name
+
+	_name="$1"; : "${_name:?}"
+	# All regress suite steps are expected to contain at least one slash.
+	[ "$_name" != "${_name#*/}" ] &&
+		[ "$(config_value "regress-${_name}-parallel")" -eq 1 ]
 }
