@@ -109,10 +109,13 @@ config_load() {
 	: "${EXECDIR:=/usr/local/libexec/robsd}"; export EXECDIR
 	PATH="${PATH}:/usr/X11R6/bin"; export PATH
 	: "${ROBSDCONFIG:=${EXECDIR}/robsd-config}"
+	: "${ROBSDEXEC:=${EXECDIR}/robsd-exec}"
 	: "${ROBSDHOOK:=${EXECDIR}/robsd-hook}"
 	: "${ROBSDLS:=${EXECDIR}/robsd-ls}"
+	: "${ROBSDREPORT:=${EXECDIR}/robsd-report}"
 	: "${ROBSDSTAT:=${EXECDIR}/robsd-stat}"
 	: "${ROBSDSTEP:=${EXECDIR}/robsd-step}"
+	: "${ROBSDWAIT:=${EXECDIR}/robsd-wait}"
 
 	_tmp="$(mktemp -t robsd.XXXXXX)"
 	{
@@ -237,7 +240,7 @@ cvs_date() {
 # date.
 cvs_log() {
 	local _date=""
-	local _id
+	local _cid
 	local _indent="  "
 	local _line
 	local _log
@@ -245,11 +248,12 @@ cvs_log() {
 	local _path
 	local _prev
 	local _repo
+	local _tmp
 	local _user
 
 	while [ $# -gt 0 ]; do
 		case "$1" in
-		-t)	shift; _tmp="${1}/cvs";;
+		-t)	shift; _tmp="$1";;
 		-c)	shift; _repo="$1";;
 		-h)	shift; _host="$1";;
 		-u)	shift; _user="$1";;
@@ -286,39 +290,39 @@ cvs_log() {
 			;;
 		date:*)
 			_date="$(cvs_field date "$_line")"
-			_id="$(cvs_field commitid "$_line")" || continue
-			if ! [ -d "${_tmp}/${_id}" ]; then
-				mkdir "${_tmp}/${_id}"
-				{
-					echo "commit ${_id}"
-					echo "Author: $(cvs_field author "$_line")"
-					echo "Date: ${_date}"
-					echo
-				} >"${_tmp}/${_id}/message"
-				# Replace trailing newline with space,
-				# simplifies sorting below.
-				echo -n "${_date} " >"${_tmp}/${_id}/date"
+			_cid="$(cvs_field commitid "$_line")" || continue
+			if ! [ -d "${_tmp}/${_cid}" ]; then
+				mkdir "${_tmp}/${_cid}"
+				cvs_field author "$_line" >"${_tmp}/${_cid}/author"
 				_message=1
 			fi
-			echo "${_indent}${_path}" >>"${_tmp}/${_id}/files"
+			echo "$_date" >>"${_tmp}/${_cid}/date"
+			echo "${_indent}${_path}" >>"${_tmp}/${_cid}/files"
 			;;
 		-[-]*|=[=]*)
 			_message=0
 			;;
 		*)
 			if [ "$_message" -eq 1 ]; then
-				echo "${_indent}${_line}" >>"${_tmp}/${_id}/message"
+				echo "${_indent}${_line}" >>"${_tmp}/${_cid}/message"
 			fi
 			;;
 		esac
 	done
 
 	# Sort each commit using the date file.
-	find "$_tmp" \( -type f -name date \) \
-		-exec sh -c 'cat $1; echo $1' _ {} \; |
+	find "$_tmp" -type f -name date |
+	while read -r _p; do
+		_date="$(sort -nr "$_p" | head -1)"
+		echo "${_date} ${_p%/*}"
+	done |
 	sort -nr |
-	sed -e 's/.* //' -e 's,/date,,' |
-	while read -r _path; do
+	while read -r _date _time _path; do
+		echo "commit ${_path##*/}"
+		echo -n "Author: "
+		cat "${_path}/author"
+		echo "Date: ${_date} ${_time}"
+		echo
 		cat "${_path}/message"
 		echo
 		cat "${_path}/files"
@@ -381,9 +385,19 @@ diff_apply() (
 #
 # Remove leftovers from cvs and patch in dir.
 diff_clean() {
-	find "$1" -type f \( \
-		-name '*.orig' -o -name '*.rej' -o -name '.#*' \) -print0 |
-	xargs -0r rm
+	local _dir
+	local _path
+
+	_dir="$1"; : "${_dir:?}"
+
+	find "$_dir" -type f \( \
+		-name '*.orig' -o -name '*.rej' -o -name '.#*' \) |
+	while read -r _path; do
+		if ! grep -sq "^/${_path##*/}/" "${_path%/*}/CVS/Entries"; then
+			echo "$_path"
+		fi
+	done |
+	xargs -r rm
 }
 
 # diff_copy -d directory dst [src ...]
@@ -415,7 +429,12 @@ diff_copy() {
 		_r="$(diff_root -d "$_root" "$_src")"
 		info "using diff ${_src} rooted in ${_r}"
 
-		{ printf '# %s\n\n' "$_src"; cat "$_src"; } >"$_dst"
+		{
+			if ! head -1 "$_src" | grep -q '^#'; then
+				printf '# %s\n\n' "$_src"
+			fi
+			cat "$_src"
+		} >"$_dst"
 		chmod 644 "$_dst"
 
 		_i=$((_i + 1))
@@ -576,6 +595,15 @@ duration_total() {
 	done
 	: "${_steps:?}"
 
+	case "$_MODE" in
+	robsd-regress)
+		regress_duration_total -s "$_steps"
+		return 0
+		;;
+	*)
+		;;
+	esac
+
 	while step_eval "$_i" "$_steps" 2>/dev/null; do
 		_i=$((_i + 1))
 
@@ -600,54 +628,6 @@ fatal() {
 	exit 1
 }
 
-# format_duration duration
-#
-# Format the given duration to a human readable representation.
-format_duration() {
-	local _d
-
-	_d="$1"; : "${_d:?}"
-
-	date -u -r "$_d" '+%T'
-}
-
-# format_size [-s] size
-#
-# Format the given size into a human readable representation.
-# Optionally include the sign if the size is a delta.
-format_size() {
-	local _abs
-	local _d=1
-	local _p=""
-	local _sign=0
-	local _size
-
-	while [ $# -gt 0 ]; do
-		case "$1" in
-		-s)	_sign=1;;
-		*)	break;;
-		esac
-		shift
-	done
-	_size="$1"; : "${_size:?}"
-
-	_abs="$(abs "$_size")"
-	if [ "$_abs" -ge "$((1024 * 1024))" ]; then
-		_d=$((1024 * 1024))
-		_p="M"
-	elif [ "$_abs" -ge "1024" ]; then
-		_d=1024
-		_p="K"
-	fi
-
-	if [ "$_sign" -eq 1 ] && [ "$_size" -ge 0 ]; then
-		echo -n '+'
-	fi
-
-	echo "${_size} ${_d} ${_p}" |
-	awk '{ printf("%.01f%s", $1 / $2, $3) }'
-}
-
 # info message ...
 #
 # Print the given message to stdout.
@@ -671,6 +651,15 @@ isempty() {
 	! find "$_path" -empty | cmp -s - /dev/null
 }
 
+# jobs_count job ...
+#
+# Get the number of running jobs
+jobs_count() (
+	# shellcheck disable=SC2068
+	set -- $@
+	echo "$#"
+)
+
 # lock_acquire root-dir build-dir
 #
 # Acquire the mutex lock.
@@ -828,382 +817,45 @@ purge() {
 	done
 }
 
-# report -b build-dir -r robsd-dir
+# report -b build-dir
 #
-# Create and save build report.
+# Create and save report.
 report() {
 	local _builddir
-	local _duration=""
-	local _exit
-	local _f
-	local _i=1
-	local _log
-	local _n
-	local _name
-	local _robsddir
 	local _report
-	local _status
 	local _steps
-	local _tmp
-	local _tmpdir
 
 	while [ $# -gt 0 ]; do
 		case "$1" in
 		-b)	shift; _builddir="$1";;
-		-r)	shift; _robsddir="$1";;
 		*)	break;;
 		esac
 		shift
 	done
-	: "${_robsddir:?}"
 	: "${_builddir:?}"
 
 	_report="${_builddir}/report"
 	_steps="$(step_path "$_builddir")"
-	_tmpdir="${_builddir}/tmp"
-	_tmp="${_tmpdir}/report"
 
 	# The steps file could be empty when a build fails to start due to
 	# another already running build.
 	[ -s "$_steps" ] || return 1
 
-	case "$_MODE" in
-	robsd-regress)
-		_status="$(regress_report_status -s "$_steps")"
-		;;
-	*)
-		# All other robsd utilities halts if a step failed, only bother
-		# checking the last non-skipped step.
-		while step_eval "-${_i}" "$_steps" 2>/dev/null; do
-			_i=$((_i + 1))
-			step_skip && continue
-			[ "$(step_value exit)" -eq 0 ] && break
-
-			_status="failed in $(step_value name)"
-			break
-		done
-		;;
-	esac
-	: "${_status:="ok"}"
-
-	if step_eval -n end "$_steps" 2>/dev/null; then
-		_duration="$(step_value duration)"
-		_duration="$(report_duration -d end -t 60 "$_duration")"
-	else
-		_duration="$(duration_total -s "$_steps")"
-		_duration="$(report_duration "$_duration")"
-	fi
-
-	# Add subject header.
-	{
-		printf 'Subject: %s: %s: ' "$_MODE" "$(hostname -s)"
-		case "$_MODE" in
-		robsd-cross)	cross_report_subject -b "$_builddir";;
-		*)		;;
-		esac
-		printf '%s\n\n' "$_status"
-	} >"$_tmp"
-
-	# Add comment to the beginning of the report.
-	if [ -e "${_builddir}/comment" ]; then
-		cat <<-EOF >>"$_tmp"
-		> comment:
-		$(cat "${_builddir}/comment")
-
-		EOF
-	fi
-
-	# Add stats to the beginning of the report.
-	{
-		cat <<-EOF
-		> stats:
-		Status: ${_status}
-		Duration: ${_duration}
-		Build: ${_builddir}
-		EOF
-
-		if [ -e  "${_builddir}/tags" ]; then
-			printf 'Tags: '
-			cat "${_builddir}/tags"
-		fi
-
-		if [ "$_MODE" = "robsd" ]; then
-			report_sizes "$(config_value bsd-reldir)"
-		fi
-	} >>"$_tmp"
-
-	_i=1
-	while step_eval "$_i" "$_steps" 2>/dev/null; do
-		_i=$((_i + 1))
-
-		step_skip && continue
-
-		_name="$(step_value name)"
-		_exit="$(step_value exit)"
-		if step_value log >/dev/null 2>&1; then
-			_log="${_builddir}/$(step_value log)"
-		else
-			_log="/dev/null"
-		fi
-
-		if [ "$_exit" -eq 0 ] &&
-		   report_skip -b "$_builddir" -n "$_name" -l "$_log" \
-			-t "$_tmpdir"
-		then
-			continue
-		fi
-
-		_duration="$(step_value duration)"
-
-		printf '\n'
-		printf '> %s:\n' "$_name"
-		printf 'Exit: %d\n' "$_exit"
-		printf 'Duration: %s\n' \
-			"$(report_duration -d "$_name" "$_duration")"
-		printf 'Log: %s\n' "$(step_value log)"
-
-		report_log -e "$_exit" -n "$_name" -l "$_log" \
-			-t "${_builddir}/tmp" >"${_builddir}/tmp/log"
-		if [ -s "${_builddir}/tmp/log" ]; then
-			trimfile "${_builddir}/tmp/log"
-			echo; cat "${_builddir}/tmp/log"
-		fi
-		rm "${_builddir}/tmp/log"
-	done >>"$_tmp"
-
-	# smtpd(8) rejects messages with carriage return not followed by a
-	# newline. Play it safe and let vis(1) encode potential carriage
-	# returns.
-	vis "$_tmp" >"$_report"
-	rm "$_tmp"
-}
-
-# report_duration [-d step] [-t threshold] duration
-#
-# Format the given duration to a human readable representation.
-# If option `-d' is given, the duration delta for the given step relative
-# to the previous successful release is also formatted if the delta is greater
-# than the given threshold.
-report_duration() {
-	local _d
-	local _delta
-	local _prev
-	local _sign
-	local _step=""
-	local _threshold=0
-
-	while [ $# -gt 0 ]; do
-		case "$1" in
-		-d)	shift; _step="$1";;
-		-t)	shift; _threshold="$1";;
-		*)	break;;
-		esac
-		shift
-	done
-	_d="$1"; : "${_d:?}"
-
-	if [ -z "$_step" ] || ! _prev="$(duration_prev "$_step")"; then
-		format_duration "$_d"
-		return 0
-	fi
-
-	_delta=$((_d - _prev))
-	if [ "$_delta" -lt 0 ]; then
-		_sign="-"
-		_delta=$((-_delta))
-	else
-		_sign="+"
-	fi
-	if [ "$_delta" -le "$_threshold" ]; then
-		format_duration "$_d"
-		return 0
-	fi
-	printf '%s (%s%s)\n' \
-		"$(format_duration "$_d")" \
-		"$_sign" \
-		"$(format_duration "$_delta")"
-}
-
-# report_log -e step-exit -n step-name -l step-log -t tmp-dir
-#
-# Writes an excerpt of the given step log.
-report_log() {
-	local _exit
-	local _f
-	local _log
-	local _name
-	local _tmpdir
-
-	case "$_MODE" in
-	robsd-ports)
-		ports_report_log "$@"
-		return $?
-		;;
-	robsd-regress)
-		regress_report_log "$@"
-		return $?
-		;;
-	*)
-		;;
-	esac
-
-	while [ $# -gt 0 ]; do
-		case "$1" in
-		-e)	shift; _exit="$1";;
-		-n)	shift; _name="$1";;
-		-l)	shift; _log="$1";;
-		-t)	shift; _tmpdir="$1";;
-		*)	break;;
-		esac
-		shift
-	done
-	: "${_exit:?}"
-	: "${_name:?}"
-	: "${_log:?}"
-	: "${_tmpdir:?}"
-
-	case "$_name" in
-	env|patch|checkflist|reboot|revert|distrib)
-		cat "$_log"
-		;;
-	cvs)
-		cvs_changelog -t "$_tmpdir"
-		;;
-	kernel)
-		tail -n 11 "$_log"
-		;;
-	*)
-		tail "$_log"
-		;;
-	esac
-}
-
-# report_size file
-#
-# If the given file is significantly larger than the same file in the previous
-# release, a human readable representation of the size and delta is reported.
-report_size() {
-	local _delta
-	local _f
-	local _name
-	local _path
-	local _prev
-	local _s1
-	local _s2
-	local _threshold
-
-	_f="$1"; : "${_f:?}"
-
-	_name="$(basename "$_f")"
-
-	[ -e "$_f" ] || return 0
-
-	_prev="$(prev_release -B | head -1)"
-	[ -z "$_prev" ] && return 0
-
-	_path="${_prev}/rel/${_name}"
-	[ -e "$_path" ] || return 0
-
-	_s1="$(ls -l "$_f" | awk '{print $5}')"
-	_s2="$(ls -l "$_path" | awk '{print $5}')"
-	_delta="$((_s1 - _s2))"
-	case "$_name" in
-	bsd.rd)	_threshold=$((1024 * 1));;
-	*)	_threshold=$((1024 * 100));;
-	esac
-	[ "$(abs "$_delta")" -ge "$_threshold" ] || return 0
-
-	echo "$_name" "$(format_size "$_s1")" \
-		"($(format_size -s "$_delta"))"
-}
-
-# report_sizes release-dir
-#
-# Report significant growth of any file present in the given release directory.
-report_sizes() {
-	local _dir
-	local _f
-	local _siz
-
-	_dir="$1"; : "${_dir:?}"
-	[ -d "$_dir" ] || return 0
-
-	find "$_dir" -type f |
-	grep -Ev '\.diff\.[0-9]+$' |
-	sort |
-	while read -r _f; do
-		_siz="$(report_size "$_f")"
-		[ -z "$_siz" ] && continue
-
-		echo "Size: ${_siz}"
-	done
-}
-
-# report_skip -b build-dir -n step-name -l step-log -t tmp-dir
-#
-# Exits zero if the given step should not be included in the report.
-report_skip() {
-	local _builddir
-	local _name
-	local _log
-
-	case "$_MODE" in
-	robsd-ports)
-		ports_report_skip "$@"
-		return $?
-		;;
-	robsd-regress)
-		regress_report_skip "$@"
-		return $?
-		;;
-	*)
-		;;
-	esac
-
-	while [ $# -gt 0 ]; do
-		case "$1" in
-		-b)	shift; _builddir="$1";;
-		-l)	shift; _log="$1";;
-		-n)	shift; _name="$1";;
-		-t)	shift;;
-		*)	break;;
-		esac
-		shift
-	done
-	: "${_builddir:?}"
-	: "${_log:?}"
-	: "${_name:?}"
-
-	case "$_name" in
-	env|reboot|dmesg|end)
-		return 0
-		;;
-	checkflist)
-		# Skip if the log only contains PS4 traces.
-		grep -vq '^\+' "$_log" || return 0
-		;;
-	patch|revert)
-		diff_list "$_builddir" "*.diff" | cmp -s - /dev/null && return 0
-		;;
-	*)
-		;;
-	esac
-
-	return 1
+	"$ROBSDREPORT" -m "$_MODE" ${ROBSDCONF:+-C ${ROBSDCONF}} "$_builddir" >"$_report"
 }
 
 # robsd -b build-dir -s step-id
 #
 # Main loop shared between utilities.
 robsd() {
-	local _exit
-	local _log
+	local _d0
+	local _d1
+	local _jobs=""
 	local _name
+	local _ncpu
 	local _s
 	local _step
 	local _steps
-	local _t0
-	local _t1
 
 	while [ $# -gt 0 ]; do
 		case "$1" in
@@ -1216,6 +868,7 @@ robsd() {
 	: "${_builddir:?}"
 	: "${_step:?}"
 
+	_ncpu="$(sysctl -n hw.ncpuonline)"
 	_steps="$(step_path "$_builddir")"
 
 	while :; do
@@ -1233,33 +886,44 @@ robsd() {
 		if [ "$_name" = "end" ]; then
 			# The duration of the end step is the accumulated
 			# duration.
+			_d1="$(duration_total -s "$_steps")"
+			_d0="$(duration_prev "$_name" || :)"
+			if [ -n "$_d0" ]; then
+				_delta="$((_d1 - _d0))"
+			else
+				_delta=0
+			fi
 			step_write -t -s "$_s" -n "$_name" -e 0 \
-				-d "$(duration_total -s "$_steps")" "$_steps"
+				-d "$_d1" -a "$_delta" "$_steps"
 			# The hook is invoked as late as possible in the exit
 			# trap handler.
 			return 0
 		fi
 
-		_log="$(log_id -b "$_builddir" -n "$_name" -s "$_s")"
-		_exit=0
-		_t0="$(date '+%s')"
-		step_write -t -l "$_log" -s "$_s" -n "$_name" -e -1 -d -1 "$_steps"
-		step_exec -f "${_builddir}/tmp/fail" -l "${_builddir}/${_log}" \
-			-s "$_name" || _exit="$?"
-		_t1="$(date '+%s')"
-		step_write -l "$_log" -s "$_s" -n "$_name" -e "$_exit" \
-			-d "$((_t1 - _t0))" "$_steps"
-		robsd_hook -v "exit=${_exit}" -v "step=${_name}"
+		if step_parallel "$_name"; then
+			# Ensure the job queue is not full.
+			if [ "$(jobs_count "$_jobs")" -eq "$_ncpu" ]; then
+				info "parallel wait $(jobs_count "${_jobs}")/${_ncpu}"
+				_jobs="$(echo "$_jobs" | xargs "$ROBSDWAIT" | xargs)"
+			fi
 
-		case "$_MODE" in
-		robsd-regress)
-			regress_step_after -b "$_builddir" -e "$_exit" \
-				-n "$_name" || return 1
-			;;
-		*)
-			[ "$_exit" -eq 0 ] || return 1
-			;;
-		esac
+			# Execute job in parallel.
+			step_exec_job -b "$_builddir" -s "$_steps" \
+				-i "$_s" -n "$_name" &
+			_jobs="${_jobs}${_jobs:+ }${!}"
+			info "parallel jobs $(jobs_count "${_jobs}")/${_ncpu}"
+		else
+			# Wait for all running jobs to finish.
+			if [ -n "$_jobs" ]; then
+				info "parallel barrier $(jobs_count "${_jobs}")/${_ncpu}"
+				echo "$_jobs" | xargs "$ROBSDWAIT" -a
+				_jobs=""
+			fi
+
+			# Execute job synchronously.
+			step_exec_job -b "$_builddir" -s "$_steps" \
+				-i "$_s" -n "$_name"
+		fi
 
 		# Reboot in progress?
 		if [ "$_name" = "reboot" ] &&
@@ -1268,7 +932,10 @@ robsd() {
 		fi
 
 		# Does robsd-kill want us dead?
-		lock_alive "$ROBSDDIR" "$_builddir" || return 1
+		if ! lock_alive "$ROBSDDIR" "$_builddir"; then
+			[ -z "$_jobs" ] || echo "$_jobs" | xargs "$ROBSDWAIT" -a
+			return 1
+		fi
 	done
 }
 
@@ -1315,10 +982,10 @@ setprogname() {
 	_PROG="$1"; export _PROG
 }
 
-# step_write [-S] [-t] [-l step-log] -s step-id -n step-name -e exit -d duration file
-#
-# Write the given step.
+# step_write [-S] [-t] [-l step-log] [-a delta]
+#            -s step-id -n step-name -e exit -d duration file
 step_write() {
+	local _delta=0
 	local _duration
 	local _exit
 	local _log=""
@@ -1330,6 +997,7 @@ step_write() {
 
 	while [ $# -gt 0 ]; do
 		case "$1" in
+		-a)	shift; _delta="$1";;
 		-S)	_skip="1";;
 		-t)	_time="$(date +%s)";;
 		-d)	shift; _duration="$1";;
@@ -1353,6 +1021,7 @@ step_write() {
 		"name=${_name}" \
 		"exit=${_exit}" \
 		"duration=${_duration}" \
+		${_delta:+delta=${_delta}} \
 		${_log:+log=${_log}} \
 		"user=${_user}" \
 		${_time:+time=${_time}} \
@@ -1367,7 +1036,7 @@ step_write() {
 step_eval() {
 	local _err=0
 	local _file
-	local _flag="-l"
+	local _flag="-i"
 	local _step
 	local _tmp
 
@@ -1387,76 +1056,114 @@ step_eval() {
 		# shellcheck disable=SC2016
 		printf '"${step}" "${name}" "${exit}" "${duration}" "${log}" '
 		# shellcheck disable=SC2016
-		printf '"${time}" "${user}" "${skip}"\n'
+		printf '"${time}" "${user}" "${skip}" "${delta}"\n'
 	} | "$ROBSDSTEP" -R -f "$_file" "$_flag" "$_step" >"$_tmp" || _err="$?"
 	[ "$_err" -eq 0 ] && eval "$(<"$_tmp")"
 	rm "$_tmp"
 	return "$_err"
 }
 
-# step_exec [-X] -f fail -l log -s step
+# step_exec [-X] -b build-dir -l log -s step
 #
 # Execute the given script and redirect any output to log.
 step_exec() (
-	local _err=0
-	local _exec
+	local _builddir
+	local _err
 	local _fail
 	local _log
-	local _robsdexec="${ROBSDEXEC:-${EXECDIR}/${_MODE}-exec}"
 	local _step
 	local _trace="yes"
 
 	while [ $# -gt 0 ]; do
 		case "$1" in
 		-X)	_trace="";;
-		-f)	shift; _fail="$1";;
+		-b)	shift; _builddir="$1";;
 		-l)	shift; _log="$1";;
 		-s)	shift; _step="$1";;
 		*)	break;;
 		esac
 		shift
 	done
-	: "${_fail:?}"
+	: "${_builddir:?}"
 	: "${_log:?}"
 	: "${_step:?}"
 
-	_exec="${EXECDIR}/${_MODE}-${_step}.sh"
-	if ! [ -e "$_exec" ]; then
-		_exec="${EXECDIR}/robsd-${_step}.sh"
-	fi
+	_fail="$(mktemp -p "${_builddir}/tmp" step-exec.XXXXXX)"
+	echo 0 >"$_fail"
 
 	[ "$DETACH" -eq 0 ] || exec >/dev/null 2>&1
 
 	{
-		if [ "$_MODE" = "robsd-regress" ] && ! [ -e "$_exec" ]; then
-			"$_robsdexec" sh -eu ${_trace:+-x} \
-				"${EXECDIR}/${_MODE}-exec.sh" "$_step" ||
-				echo "$?" >"$_fail"
+		"$ROBSDEXEC" -m "$_MODE" ${ROBSDCONF:+"-C${ROBSDCONF}"} \
+			${_trace:+-x} "$_step" || echo "$?" >"$_fail"
 
+		if [ "$_MODE" = "robsd-regress" ]; then
 			# Regress tests can fail but still exit zero, check the
 			# log for failures.
 			regress_failed "$_log" && echo 1 >"$_fail"
-		else
-			"$_robsdexec" sh -eu ${_trace:+-x} "$_exec" ||
-				echo "$?" >"$_fail"
 		fi
 	} </dev/null 2>&1 | tee "$_log"
-	if [ -e "$_fail" ]; then
-		_err="$(<"$_fail")"
-		rm -f "$_fail"
-	fi
+	_err="$(<"$_fail")"
+	rm -f "$_fail"
 	return "$_err"
 )
 
-# step_failures file
+# step_exec_job -b build-dir -s steps -i step-id -n step-name
 #
-# Get the number of failing steps.
-step_failures() {
-	local _file
+# Execute a single step.
+step_exec_job() {
+	local _builddir
+	local _d0
+	local _d1
+	local _delta
+	local _exit=0
+	local _id
+	local _log
+	local _name
+	local _steps
+	local _t0
+	local _t1
 
-	_file="$1"; : "${_file:?}"
+	while [ $# -gt 0 ]; do
+		case "$1" in
+		-b)	shift; _builddir="$1";;
+		-s)	shift; _steps="$1";;
+		-i)	shift; _id="$1";;
+		-n)	shift; _name="$1";;
+		*)	break;;
+		esac
+		shift
+	done
+	: "${_builddir:?}"
+	: "${_id:?}"
+	: "${_name:?}"
+	: "${_steps:?}"
 
-	tail -n +2 "$_file" | awk -F , '{s+= $3 == 0 ? 0 : 1} END {print s}'
+	_log="$(log_id -b "$_builddir" -n "$_name" -s "$_id")"
+	_t0="$(date '+%s')"
+	step_write -t -l "$_log" -s "$_id" -n "$_name" -e -1 -d -1 "$_steps"
+	step_exec -b "$_builddir" -l "${_builddir}/${_log}" -s "$_name" || _exit="$?"
+	_t1="$(date '+%s')"
+	_d1="$((_t1 - _t0))"
+	_d0="$(duration_prev "$_name" || :)"
+	if [ -n "$_d0" ]; then
+		_delta="$((_d1 - _d0))"
+	else
+		_delta=0
+	fi
+	step_write -l "$_log" -s "$_id" -n "$_name" -e "$_exit" -d "$_d1" \
+		-a "$_delta" "$_steps"
+
+	robsd_hook -v "exit=${_exit}" -v "step=${_name}"
+
+	case "$_MODE" in
+	robsd-regress)
+		regress_step_after -b "$_builddir" -e "$_exit" -n "$_name" || return 1
+		;;
+	*)
+		[ "$_exit" -eq 0 ] || return 1
+		;;
+	esac
 }
 
 # step_field step-name
@@ -1476,6 +1183,7 @@ step_field() {
 	time)		echo 5;;
 	user)		echo 6;;
 	skip)		echo 7;;
+	delta)		echo 8;;
 	*)		echo -1;;
 	esac
 }
@@ -1525,43 +1233,8 @@ step_path() {
 # steps
 #
 # Get the names of all steps in execution order.
-# The last step named end is a sentinel step without a corresponding step
-# script.
 steps() {
-	case "$_MODE" in
-	robsd)
-		cat <<-EOF
-		env
-		cvs
-		patch
-		kernel
-		reboot
-		env
-		base
-		release
-		checkflist
-		xbase
-		xrelease
-		image
-		hash
-		revert
-		distrib
-		dmesg
-		end
-		EOF
-		;;
-	robsd-cross)
-		cross_steps
-		;;
-	robsd-ports)
-		ports_steps
-		;;
-	robsd-regress)
-		regress_steps
-		;;
-	*)
-		;;
-	esac
+	"$ROBSDSTEP" -L -m "$_MODE" ${ROBSDCONF:+"-C${ROBSDCONF}"}
 }
 
 # step_next file
@@ -1601,6 +1274,20 @@ step_next() {
 	return 1
 }
 
+# step_parallel step-name
+#
+# Exits zero if the step can be executed in parallel.
+step_parallel() {
+	case "$_MODE" in
+	robsd-regress)
+		regress_step_parallel "$_name"
+		;;
+	*)
+		return 1
+		;;
+	esac
+}
+
 # step_skip
 #
 # Exits zero if the step has been skipped.
@@ -1662,7 +1349,7 @@ trap_exit() {
 	if [ "$_err" -ne 0 ] ||
 	   step_eval -n end "$_steps" 2>/dev/null
 	then
-		if report -r "$_robsddir" -b "$_builddir" &&
+		if report -b "$_builddir" &&
 		   [ "$DETACH" -ne 0 ]; then
 			# Do not send mail during interactive invocations.
 			sendmail root <"${_builddir}/report"
@@ -1681,18 +1368,6 @@ trap_exit() {
 	return "$_err"
 }
 
-# trimfile path
-#
-# Remove empty lines at the end of the given file.
-trimfile() {
-	local _path
-
-	_path="$1"; : "${_path:?}"
-	while [ "$(sed -n -e '$p' "$_path")" = "" ]; do
-		sed -i -e '$d' "$_path"
-	done
-}
-
 # unpriv [-c class] user utility argument ...
 # unpriv [-c class] user
 #
